/* Special .init and .fini section support.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   In addition to the permissions in the GNU Library General Public
   License, the Free Software Foundation gives you unlimited
   permission to link the compiled version of this file with other
   programs, and to distribute those programs without any restriction
   coming from the use of this file.  (The Library General Public
   License restrictions do apply in other respects; for example, they
   cover modification of the file, and distribution when not linked
   into another program.)

   The GNU C Library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* This file is compiled into assembly code which is then munged by a sed
   script into two files: crti.s and crtn.s.

   * crti.s puts a function prologue at the beginning of the
   .init and .fini sections and defines global symbols for
   those addresses, so they can be called as functions.

   * crtn.s puts the corresponding function epilogues
   in the .init and .fini sections. */

/* XXX hack alert -- the initfini stunt doesn't compile for the i370
 * assembler because the literal pool is not normally dumped in the 
 * prologue, and will result in missing address const assembler errors.
 * We hack around this by forcing a literal pool to be dumped at the 
 * end of the prologues.
 */
#define I370 1

#include <stdlib.h>

/* We use embedded asm for .section unconditionally, as this makes it
   easier to insert the necessary directives into crtn.S. */
#define SECTION(x) asm (".section " x );

/* Embed an #include to pull in the alignment and .end directives. */
asm ("\n#include \"defs.h\"");

/* The initial common code ends here. */
asm ("\n/*@HEADER_ENDS*/");

/* To determine whether we need .end and .align: */
asm ("\n/*@TESTS_BEGIN*/");
void
dummy (void (*foo) (void))
{
  if (foo)
    (*foo) ();
}
asm ("\n/*@TESTS_END*/");

/* The beginning of _init:  */
asm ("\n/*@_init_PROLOG_BEGINS*/");

SECTION (".init")
void
_init (void)
{
  /* We cannot use the normal constructor mechanism in gcrt1.o because it
     appears before crtbegin.o in the link, so the header elt of .ctors
     would come after the elt for __gmon_start__.  One approach is for
     gcrt1.o to reference a symbol which would be defined by some library
     module which has a constructor; but then user code's constructors
     would come first, and not be profiled.  */
  extern void __gmon_start__ (void) __attribute__ ((weak)); /*weak_extern (__gmon_start__);*/

#ifndef WEAK_GMON_START
  __gmon_start__ ();
#else
  if (__gmon_start__)
    __gmon_start__ ();
#endif

#ifdef I370
  asm (
	".LPOOL1:;		"
	".ltorg;		"
	".drop r3;		"
	"	.balign 4;	"
	".LPGT1: ;		"
        " 	.long   .LPG1;	"
      );
#endif

  asm ("ALIGN");
  asm("END_INIT");
  asm ("\n/*@_init_PROLOG_ENDS*/");

  /* Now the epilog. */
  asm ("\n/*@_init_EPILOG_BEGINS*/");
  SECTION(".init");
#ifdef I370
  asm (
	".data ;		"
	".using	.,r12 ;		"
	".LPG1: ;		"
      );
#endif
}
asm ("END_INIT");
#ifdef I370
  asm (
	".drop	r12 ;		"
      );
#endif
#ifndef WEAK_GMON_START
SECTION(".text");

/* This version of __gmon_start__ is used if no other is found.  By providing
   a default function we avoid the need to test whether the pointer is NULL,
   which can be painful on some machines.  */

void __attribute__ ((weak))
__gmon_start__ (void)
{
  /* do nothing */
}
#endif

/* End of the _init epilog, beginning of the _fini prolog. */
asm ("\n/*@_init_EPILOG_ENDS*/");
asm ("\n/*@_fini_PROLOG_BEGINS*/");

SECTION (".fini")
void
_fini (void)
{

#ifdef I370
  asm (
	".LPOOL3:;		"
	".ltorg;		"
	"	.balign 4;	"
	".LPGT3: ;		"
        "	.long   .LPG3;	"
      );
#endif

  /* End of the _fini prolog. */
  asm ("ALIGN");
  asm ("END_FINI");
  asm ("\n/*@_fini_PROLOG_ENDS*/");

  {
    /* Let GCC know that _fini is not a leaf function by having a dummy
       function call here.  We arrange for this call to be omitted from
       either crt file.  */
    extern void i_am_not_a_leaf (void);
    i_am_not_a_leaf ();
  }

  /* Beginning of the _fini epilog. */
  asm ("\n/*@_fini_EPILOG_BEGINS*/");
  SECTION (".fini");
#ifdef I370
  asm (
	".data ;		"
	".using .,r12 ;		"
	".LPG3: ;		"
      );
#endif
}
asm ("END_FINI");

/* End of the _fini epilog.  Any further generated assembly (e.g. .ident)
   is shared between both crt files. */
asm ("\n/*@_fini_EPILOG_ENDS*/");
asm ("\n/*@TRAILER_BEGINS*/");

/* End of file. */
