/* Copyright (C) 1992, 1997, 1998 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <sysdeps/unix/sysdep.h>

/* For Linux we can use the system call table in the header file
	/usr/include/asm/unistd.h
   of the kernel.  But these symbols do not follow the SYS_* syntax
   so we have to redefine the `SYS_ify' macro here.  */
#undef SYS_ify
#ifdef __STDC__
# define SYS_ify(syscall_name)	__NR_##syscall_name
#else
# define SYS_ify(syscall_name)	__NR_/**/syscall_name
#endif

#if defined(__ASSEMBLER__) || defined(ASSEMBLER)
/* ret is bogus */
#define ret

/* Linux uses a negative return value to indicate syscall errors,
   unlike most Unices, which use the condition codes' carry flag.

   Since version 2.1 the return value of a system call might be
   negative even if the call succeeded.  E.g., the `lseek' system call
   might return a large offset.  Therefore we must not anymore test
   for < 0, but test for a real error by making sure the value in R0
   is a real error number.  Linus said he will make sure the no syscall
   returns a value in -1 .. -4095 as a valid result so we can savely
   test with -4095. 
 */
/* Linux takes system call args in registers:
        syscall number  r1
        arg 1           r5
        arg 2           r6
        arg 3           r7
        arg 4           r8
        arg 5           r9
*/
/* XXX hack alert --  the implementation below uses the
 * standard prologue but we could save a few cycles by not 
 * buying the stack frame.  Not that big a deal, it handy 
 * for debuggng at the moment.
 */

#define DOARGS_0 /* nothing */
#define DOARGS_1 DOARGS_0 ; L	r5,88(r13)
#define DOARGS_2 DOARGS_1 ; L	r6,92(r13)
#define DOARGS_3 DOARGS_2 ; L	r7,96(r13)
#define DOARGS_4 DOARGS_3 ; L	r8,100(r13)
#define DOARGS_5 DOARGS_4 ; L	r9,104(r13)
#define DOARGS_6 DOARGS_5 ; L	r10,108(r13)

#undef PSEUDO
#define PSEUDO(name, syscall_name, args)	\
.text						;\
        .balign 4				;\
.globl name					;\
        .type    name,@function			;\
name:						;\
/* Function prologue */				;\
/* args=0x20 frame_size=0 aligned size=0x80 */	;\
        .using  .,r15				;\
        B       1f				;\
        .long   128				;\
        .long   .LPGT0				;\
        .drop	r15				;\
        .balign 2				;\
1:						;\
        STM     r13,r12,8(r11)			;\
	LR	r13,r11				;\
        A	r11,4(,r15)			;\
        L	r4,8(,r15)			;\
        BASR    r3,0				;\
        .using  .,r3				;\
.LPG0:						;\
        ST	r4,0(,r13)			;\
/* Function code */				;\
	LA	r1,__NR_##syscall_name(0,0)	;\
	DOARGS_##args				;\
	SVC	0				;\
        LTR     r15,r15				;\
	BNL	2f				;\
        L       r0,=F'-4095'			;\
        CR      r15,r0				;\
	BL	2f				;\
        LCR     r6,r15				;\
        L       r15,=V(__errno_location)	;\
        BASR    lr,r15				;\
        ST      r6,0(,r15)			;\
        L       r15,=F'-1'			;\
2:						;\
/* Function epilogue */				;\
        L       r14,12(,r13)			;\
        LM      r2,r12,28(r13)			;\
        L       r13,8(,r13)			;\
        BASR    r1,r14				;\
/* Function literal pool */			;\
        .balign 4				;\
        .ltorg					;\
/* Function page table */			;\
        .balign 4				;\
.LPGT0:						;\
        .long   .LPG0				;\
.Lfe1:						;\
        .size    name,.Lfe1-name		;\



#else /* not __ASSEMBLER__ */
extern int * __errno_location(void);

/* Define a macro which expands into the inline wrapper code for a system
   call.  */
#undef INLINE_SYSCALL
#define INLINE_SYSCALL(name, nr, args...)                       \
  ({ unsigned int _sys_result;                                  \
     {                                                          \
       register int _r1 asm ("r1");                             \
       register int _r15 asm ("r15");                           \
       LOAD_ARGS_##nr (args)                                    \
       _r1 = SYS_ify(name);					\
       asm volatile ("SVC	0"                              \
                     : "=r" (_r15)                              \
                     : "r" (_r1) ASM_ARGS_##nr  		\
                     : CLOBBER_##nr);                           \
       _sys_result = _r15;                                      \
     }                                                          \
     if (_sys_result >= (unsigned int) -4095)                   \
       {                                                        \
         *(__errno_location()) = -_sys_result;                  \
         _sys_result = (unsigned int) -1;                       \
       }                                                        \
     (int) _sys_result; })

#define LOAD_ARGS_0()
#define ASM_ARGS_0
#define CLOBBER_0	"r1","r15"

#define LOAD_ARGS_1(r5)                         \
  register int _r5 asm ("r5") = (int) (r5);     \
  LOAD_ARGS_0 ()
#define ASM_ARGS_1      ASM_ARGS_0, "r" (_r5)
#define CLOBBER_1	CLOBBER_0,"r5"

#define LOAD_ARGS_2(r5, r6)                     \
  register int _r6 asm ("r6") = (int) (r6);     \
  LOAD_ARGS_1 (r5)
#define ASM_ARGS_2      ASM_ARGS_1, "r" (_r6)
#define CLOBBER_2	CLOBBER_1,"r6"

#define LOAD_ARGS_3(r5, r6, r7)                 \
  register int _r7 asm ("r7") = (int) (r7);     \
  LOAD_ARGS_2 (r5, r6)
#define ASM_ARGS_3      ASM_ARGS_2, "r" (_r7)
#define CLOBBER_3	CLOBBER_2,"r7"

#define LOAD_ARGS_4(r5, r6, r7, r8)             \
  register int _r8 asm ("r8") = (int) (r8);     \
  LOAD_ARGS_3 (r5, r6, r7)
#define ASM_ARGS_4      ASM_ARGS_3, "r" (_r8)
#define CLOBBER_4	CLOBBER_3,"r8"

#define LOAD_ARGS_5(r5, r6, r7, r8, r9)         \
  register int _r9 asm ("r9") = (int) (r9);     \
  LOAD_ARGS_4 (r5, r6, r7, r8)
#define ASM_ARGS_5      ASM_ARGS_4, "r" (_r9)
#define CLOBBER_5	CLOBBER_4,"r9"

#define LOAD_ARGS_6(r5, r6, r7, r8, r9, r10)    \
  register int _r10 asm ("r10") = (int) (r10);  \
  LOAD_ARGS_5 (r5, r6, r7, r8, r9)
#define ASM_ARGS_6      ASM_ARGS_5, "r" (_r10)
#define CLOBBER_6	CLOBBER_5,"r10"

#endif /* __ASSEMBLER__ */
